#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Extract Tokens using a Sliding Window Algorithm """


from baseblock import Enforcer
from baseblock import Stopwatch
from baseblock import BaseObject


class SlidingWindowExtract(BaseObject):
    """ Extract Tokens using a Sliding Window Algorithm """

    def __init__(self):
        """ Change Log

        Created:
            24-Aug-2022
            craigtrim@gmail.com
        Updated:
            26-Aug-2022
            craigtrim@gmail.com
            *   ensure consistency of output
                https://github.com/craigtrim/dbpedia-tag/issues/1
        """
        BaseObject.__init__(self, __name__)

    @staticmethod
    def _normalize(results: list) -> list:
        normalized = []
        for result in results:

            output = result
            if type(result) == list:
                output = ' '.join(result).strip()

            normalized.append(output.replace('  ', ' ').strip())

        return normalized

    def _process(self,
                 gram_size: int,
                 candidates: list) -> list:
        results = []

        if gram_size == len(candidates):
            return [' '.join(candidates).strip()]

        if gram_size == 1:
            return self._normalize(candidates)

        x = 0
        y = x + gram_size

        while y <= len(candidates):
            results.append(candidates[x: y])

            x += 1
            y = x + gram_size

        normalized = []
        for result in results:
            output = ' '.join(result).strip()
            output = output.replace('  ', ' ')
            normalized.append(output)

        return normalized

    def process(self,
                gram_size: int,
                candidates: list) -> list:
        """ Entry Point

        Args:
            gram_size (int): the gram size to execute the sliding window at
            candidates (list): the incoming tokens to run the sliding window against

        Returns:
            list: extraction of candidates by gram level
        """
        sw = Stopwatch()

        results = self._process(gram_size=gram_size,
                                candidates=candidates)

        if self.isEnabledForDebug:
            if results:
                Enforcer.is_list_of_str(results)

            self.logger.debug('\n'.join([
                "Sliding Window Extraction Completed",
                f"\tGram Size: {gram_size}",
                f"\tExtracted Tokens: {results}",
                f"\tTotal Time: {str(sw)}"]))

        return results



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Perform Sliding Window Extraction for Candidate Synonym Swapping """


from baseblock import Stopwatch
from baseblock import BaseObject
from baseblock import TextMatcher

from fast_sentence_tokenize import Tokenizer


class PerformTextReplacement(BaseObject):
    """ Perform Sliding Window Extraction for Candidate Synonym Swapping """

    def __init__(self,
                 original_text: str):
        """ Change Log:

        Created:
            26-Aug-2022
            craigtrim@gmail.com
            *   refactor sliding window algorithm
                https://github.com/craigtrim/dbpedia-tag/issues/2
        """
        BaseObject.__init__(self, __name__)
        self._tokenize = Tokenizer().input_text
        self._original_text = original_text
        self._d_coords = {}
        self._d_canons = {}

    def coords(self) -> dict:
        return self._d_coords

    def canons(self) -> dict:
        return self._d_canons

    # TODO: baseblock.TextMatcher (0.1.14+)

    @staticmethod
    def replace(input_text: str,
                old_value: str,
                new_value: str,
                case_sensitive: bool = False,
                recursive: bool = False) -> bool:
        """ Check for the existence of a value in input text

        Args:
            input_text (str): any text string to search in
            old_value (str): the value that must exist in the input text
            new_value (str): the value that will replace 'old value' within the input text
            case_sensitive (bool): True if case sensitivity matters
            recursive (bool): if True, then apply method recursively until all changes are made

        Returns:
            bool: True if the value exists in the input text
        """

        if not case_sensitive:
            old_value = old_value.lower()
            input_text = input_text.lower()

        if old_value == input_text:
            return new_value

        original_text = input_text

        match_lr = f" {old_value} "
        if match_lr in input_text:
            input_text = input_text.replace(match_lr, f" {new_value} ")

        match_l = f" {old_value}"
        if input_text.endswith(match_l):
            input_text = input_text.replace(match_lr, f" {new_value}")

        match_r = f"{old_value} "
        if input_text.startswith(match_r):
            input_text = input_text.replace(match_lr, f"{new_value} ")

        if recursive and original_text != input_text:
            return TextMatcher.replace(
                input_text=input_text,
                old_value=old_value,
                new_value=new_value,
                case_sensitive=case_sensitive)

        return input_text

    @staticmethod
    def get_coords(value: str,
                   input_text: str,
                   case_sensitive: bool = False) -> bool:
        """ Find the X,Y coords (if any) for a given value in the supplied input text

        Args:
            input_text (str): any text string to search in
            value (str): the value to find coords for
            case_sensitive (bool): True if case sensitivity matters

        Returns:
            bool: True if the value exists in the input text
        """

        if not case_sensitive:
            value = value.lower()
            input_text = input_text.lower()

        if value == input_text:
            return 0, len(value)

        match_lr = f" {value} "
        if match_lr in input_text:
            x = input_text.index(match_lr)
            return x, x + len(value)

        match_l = f" {value}"
        if input_text.endswith(match_l):
            return len(input_text) - len(value), len(input_text)

        match_r = f"{value} "
        if input_text.startswith(match_r):
            return 0, len(value)

        # not found
        return None, None

    def _process(self,
                 input_text: str,
                 d_entity_map: dict) -> dict:

        for original_entity in d_entity_map:

            canon = d_entity_map[original_entity]
            key = f"entity_{canon.replace(' ', '_').lower()}"

            x, y = self.get_coords(value=original_entity,
                                   input_text=input_text,
                                   case_sensitive=False)

            if x is None or y is None:
                continue

            self._d_coords[key] = [x, y]
            self._d_canons[key] = original_entity

            input_text = self.replace(input_text=input_text,
                                      old_value=original_entity,
                                      new_value=key,
                                      case_sensitive=False,
                                      recursive=False)

        return input_text

    def process(self,
                d_entity_map: dict,
                input_text: str) -> str:
        """ Entry Point

        Args:
            d_entity_map (dict): maps original form to canonical form
                e.g., 
                {
                    'a&w restaurant': 'a&w_restaurants'
                }
            input_text (str): the input text in which original forms will be replaced with canonical forms
                e.g.,       "let's go to the a&w restaurant"
                becomes     "let's go to the entity_a&w_restaurants"          

        Returns:
            str: the modified text
        """

        sw = Stopwatch()

        output_text = self._process(input_text=input_text,
                                    d_entity_map=d_entity_map)

        if self.isEnabledForInfo:
            self.logger.info('\n'.join([
                "Text Replacement Completed",
                f"\tInput Text: {input_text}",
                f"\tOutput Text: {output_text}",
                f"\tTotal Time: {str(sw)}"]))

        return output_text



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Match Tokens against the dbPedia Dictionaries """


from typing import Callable

from baseblock import Enforcer
from baseblock import Stopwatch
from baseblock import BaseObject


class SlidingWindowMatch(BaseObject):
    """ Match Tokens against the dbPedia Dictionaries """

    def __init__(self,
                 entity_exists: Callable):
        """ Change Log

        Created:
            24-Aug-2022
            craigtrim@gmail.com
        Updated:
            26-Aug-2022
            craigtrim@gmail.com

        Args:
            entity_exists (Callable): callback to the dbpedia-ent function
        """
        BaseObject.__init__(self, __name__)
        self._entity_exists = entity_exists

    def _process(self,
                 candidates: list) -> list:

        normalized = []

        for candidate in candidates:
            if self._entity_exists(candidate):
                normalized.append(candidate)

        return normalized

    def process(self,
                candidates: list) -> list:
        sw = Stopwatch()

        results = self._process(candidates)

        if self.isEnabledForDebug:
            if results and len(results):
                Enforcer.is_list_of_str(results)
            removed = set(candidates).difference(set(results))

            self.logger.debug('\n'.join([
                "Sliding Window Location Completed",
                f"\tRemoved Tokens: {list(removed)}",
                f"\tRetained Tokens: {results}",
                f"\tTotal Time: {str(sw)}"]))

        return results




#!/usr/bin/env python
# -*- coding: UTF-8 -*-


from dataclasses import replace
from pprint import pprint

from baseblock import Enforcer
from baseblock import Stopwatch
from baseblock import BaseObject
from baseblock import ServiceEventGenerator

from fast_sentence_tokenize import Tokenizer

from dbpedia_ent.bp import Finder
from dbpedia_tag.svc import FindExactMatches
from dbpedia_tag.svc import FindFuzzyMatches
from dbpedia_tag.svc import PerformTextReplacement


class Tagger(BaseObject):
    """ Orchestrate Taxonomy Generation """

    def __init__(self):
        """ Change Log:

        Created:
            24-Aug-2022
            craigtrim@gmail.com
        """
        BaseObject.__init__(self, __name__)
        finder = Finder()
        self._entity_exists = finder.exists
        self._tokenize = Tokenizer().input_text

        self._generate_event = ServiceEventGenerator().process
        self._fuzzy_matcher = FindFuzzyMatches(finder.find).process
        self._exact_matcher = FindExactMatches(self._entity_exists).process

    def _process(self,
                 input_text: str,
                 original_text: str):

        replace_text = PerformTextReplacement(original_text)

        tokens = [x.strip() for x in self._tokenize(input_text)]
        tokens = [x for x in tokens if not x.startswith('entity_')]

        for i in reversed(range(1, 4)):

            d_match_results = self._exact_matcher(
                gram_size=i, tokens=tokens)
            if not d_match_results:
                continue

            if d_match_results['matched']:
                # need to induce a structure to fit the service
                d_exact = {k: k for k in d_match_results['matched']}
                input_text = replace_text.process(
                    d_entity_map=d_exact,
                    input_text=input_text)

            if d_match_results['unmatched']:
                d_fuzzy = self._fuzzy_matcher(
                    d_match_results['unmatched'])

                d_fuzzy = {k: d_fuzzy[k] for k in d_fuzzy if d_fuzzy[k]}

                input_text = replace_text.process(
                    d_entity_map=d_fuzzy,
                    input_text=input_text)

            tokens = [x.strip() for x in self._tokenize(input_text)]
            tokens = [x for x in tokens if not x.startswith('entity_')]

        return {
            'input_text': original_text,
            'output_text': input_text,
            'coords': replace_text.coords(),
            'canons': replace_text.canons()
        }

    def process(self,
                input_text: str):

        sw = Stopwatch()

        output_events = []
        d_result = self._process(
            input_text=input_text.lower().strip(),
            original_text=input_text)

        output_events.append(self._generate_event(
            service_name=self.component_name(),
            event_name='tagger',
            stopwatch=sw,
            data=d_result))

        return {
            'text': input_text,
            'events': output_events
        }



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Perform Sliding Window Extraction for Candidate Synonym Swapping """


from typing import Callable

from baseblock import Enforcer
from baseblock import Stopwatch
from baseblock import BaseObject

from dbpedia_tag.dmo import SlidingWindowExtract
from dbpedia_tag.dmo import SlidingWindowBlacklist
from dbpedia_tag.dmo import SlidingWindowMatch


class FindExactMatches(BaseObject):
    """ Perform Sliding Window Extraction for Candidate Synonym Swapping """

    def __init__(self,
                 entity_exists: Callable):
        """ Change Log:

        Created:
            24-Aug-2022
            craigtrim@gmail.com
        Updated:
            26-Aug-2022
            craigtrim@gmail.com
            *   refactor sliding window algorithm
                https://github.com/craigtrim/dbpedia-tag/issues/1

        Args:
            entity_exists (Callable): _description_
        """
        BaseObject.__init__(self, __name__)
        self._entity_exists = entity_exists

        self._extract_candidates = SlidingWindowExtract().process
        self._blacklist_candidates = SlidingWindowBlacklist().process
        self._match_candidates = SlidingWindowMatch(
            self._entity_exists).process

    def _process(self,
                 gram_size: int,
                 tokens: list) -> dict:

        candidates = self._extract_candidates(
            candidates=tokens,
            gram_size=gram_size)

        if not candidates or not len(candidates):
            return None

        candidates = self._blacklist_candidates(candidates)
        matched = self._match_candidates(candidates)

        unmatched = [x for x in candidates if x not in matched]

        return {
            'input': tokens,
            'candidates': candidates,
            'matched': matched,
            'unmatched': unmatched
        }

    def process(self,
                gram_size: int,
                tokens: list) -> list:

        sw = Stopwatch()

        d_results = self._process(tokens=tokens,
                                  gram_size=gram_size)

        if self.isEnabledForInfo:

            def total_results() -> int:
                if d_results:
                    return len(d_results)
                return 0

            if self.isEnabledForInfo:
                self.logger.info('\n'.join([
                    "Sliding Window Completed",
                    f"\tGram Size: {gram_size}",
                    f"\tTotal Results: {total_results()}",
                    f"\tTotal Time: {str(sw)}"]))

        if self.isEnabledForDebug and d_results:
            Enforcer.keys(d_results,
                          'input', 'candidates', 'matched', 'unmatched')

        return d_results



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dbpedia_tag',
 'dbpedia_tag.bp',
 'dbpedia_tag.dmo',
 'dbpedia_tag.dto',
 'dbpedia_tag.svc']

package_data = \
{'': ['*']}

install_requires = \
['baseblock', 'dbpedia-ent', 'fast-sentence-tokenize']

setup_kwargs = {
    'name': 'dbpedia-tag',
    'version': '0.1.10',
    'description': 'Perform runtime tagging and annotation of unstructured text',
    'long_description': None,
    'author': 'Craig Trim',
    'author_email': 'craigtrim@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '==3.8.5',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Blacklist certain Tokens from the sliding window result """


from baseblock import TextUtils
from baseblock import Enforcer
from baseblock import Stopwatch
from baseblock import BaseObject

from dbpedia_tag.dto import stopwords


class SlidingWindowBlacklist(BaseObject):
    """ Blacklist certain Tokens from the sliding window result """

    def __init__(self):
        """
        Created:
            24-Aug-2022
            craigtrim@gmail.com
        Updated:
            26-Aug-2022
            craigtrim@gmail.com
            *   refactor filtering into multiple methods
        """
        BaseObject.__init__(self, __name__)

    def _length_check(self,
                      candidates: list) -> list:

        normalized = []
        for candidate in candidates:

            len_of_candidate = len(candidate)
            if len_of_candidate < 5:
                continue

            # count the length minus the spaces
            if len_of_candidate - candidate.count(' ') < 5:
                continue

            normalized.append(candidate)

        return normalized

    def _punkt_check(self,
                     candidates: list) -> list:

        normalized = []
        for candidate in candidates:

            if ' ' not in candidate:
                normalized.append(candidate)
                continue

            # avoid any n-grams where a single gram is only punctuation
            grams = candidate.split()
            if not len([x for x in grams if TextUtils.is_punctuation(x)]):
                normalized.append(candidate)

        return normalized

    def _stopword_filter(self,
                         candidates: list) -> list:
        """ Stopword Filtering removes low-value phrases and words

        Architectural Decisions:
            https://github.com/craigtrim/dbpedia-tag/issues/3
            Net Net: We can't remove all low-value words; this has to be a consumer strategy

        Args:
            candidates (list): the incoming candidates

        Returns:
            list: the potentially filtered candidates
        """

        normalized = []
        for candidate in candidates:

            if candidate in stopwords:
                continue

            # avoid any n-grams where even a single gram is a stopwrd
            grams = candidate.split()
            if not len([x for x in grams if x in stopwords]):
                normalized.append(candidate)

        return normalized

    def _process(self,
                 candidates: list) -> list:

        candidates = self._stopword_filter(candidates)
        candidates = self._length_check(candidates)
        candidates = self._punkt_check(candidates)

        return candidates

    def process(self,
                candidates: list) -> list:
        sw = Stopwatch()

        results = self._process(candidates)

        if self.isEnabledForDebug:
            if results:
                Enforcer.is_list_of_str(results)

            removed = set(candidates).difference(set(results))

            self.logger.debug('\n'.join([
                "Sliding Window Blacklist Completed",
                f"\tRemoved Tokens: {list(removed)}",
                f"\tRetained Tokens: {results}",
                f"\tTotal Time: {str(sw)}"]))

        return results
